import os
import pytest
from unittest.mock import patch
from typing import Dict, Any
from versionhq.agent.model import Agent
from versionhq.task.model import Task, ResponseField, TaskOutput, ConditionalTask
from versionhq.tool.model import Tool, ToolSet

DEFAULT_MODEL_NAME = os.environ.get("LITELLM_MODEL_NAME", "gpt-3.5-turbo")
LITELLM_API_KEY = os.environ.get("LITELLM_API_KEY")


def test_sync_execute_task():
    agent = Agent(
        role="demo agent 1",
        goal="My amazing goals",
        backstory="My amazing backstory",
        verbose=True,
        llm=DEFAULT_MODEL_NAME,
        max_tokens=3000,
    )

    task = Task(
        description="Analyze the client's business model and define the optimal cohort timeframe.",
        expected_output_json=True,
        expected_output_pydantic=True,
        output_field_list=[
            ResponseField(title="test1", type=str, required=True),
            ResponseField(title="test2", type=list, required=True),
        ],
    )
    res = task.execute_sync(agent=agent)

    assert isinstance(res, TaskOutput)
    assert res.task_id is task.id
    assert res.raw is not None
    assert isinstance(res.raw, str)
    assert res.json_dict is not None
    assert isinstance(res.json_dict, dict)
    assert res.pydantic is not None

    if hasattr(res.pydantic, "output"):
        assert res.pydantic.output is not None
    else:
        assert hasattr(res.pydantic, "test1")
        if res.pydantic.test1:
            assert type(res.pydantic.test1) == str

        assert hasattr(res.pydantic, "test2")
        if res.pydantic.test2:
            assert type(res.pydantic.test2) == list


def test_async_execute_task():
    agent = Agent(role="demo agent", goal="My amazing goals")
    task = Task(description="Return string: 'test'")

    with patch.object(Agent, "execute_task", return_value="test") as execute:
        execution = task.execute_async(agent=agent)
        result = execution.result()
        assert result.raw == "test"
        execute.assert_called_once_with(task=task, context=None)


def test_sync_execute_with_task_context():
    """
    Use case = One agent handling multiple tasks sequentially using context set in the main task.
    """

    agent = Agent(role="demo 3", goal="My amazing goals", verbose=True, llm=DEFAULT_MODEL_NAME, max_tokens=3000,)
    sub_task = Task(
        description="analyze the client's business model",
        output_field_list=[
            ResponseField(title="subtask_result", type=str, required=True),
        ]
    )
    main_task = Task(
        description="Define the optimal cohort timeframe in days and target audience.",
        output_field_list=[
            ResponseField(title="test1", type=int, required=True),
            ResponseField(title="test2", type=str, required=True),
        ],
        context=[sub_task,]
    )
    res = main_task.execute_sync(agent=agent)

    assert isinstance(res, TaskOutput)
    assert res.task_id is main_task.id
    assert res.raw is not None
    assert isinstance(res.raw, str)
    assert res.json_dict is not None
    assert isinstance(res.json_dict, dict)
    assert res.pydantic is None
    assert sub_task.output is not None
    assert sub_task.output.json_dict is not None
    assert "subtask_result" in main_task.prompt()


def test_sync_execute_task_with_prompt_context():
    """
    Use case:
    - One agent handling multiple tasks sequentially using context set in the main task.
    - On top of that, the agent receives context when they execute the task.
    """

    agent = Agent(
        role="demo agent 4",
        goal="My amazing goals",
        backstory="My amazing backstory",
        verbose=True,
        llm=DEFAULT_MODEL_NAME,
        max_tokens=3000,
    )

    sub_task = Task(
        description="analyze the client's business model",
        expected_output_json=True,
        expected_output_pydantic=False,
        output_field_list=[
            ResponseField(title="result", type=str, required=True),
        ]
    )
    main_task = Task(
        description="Define the optimal cohort timeframe in days and target audience.",
        expected_output_json=True,
        expected_output_pydantic=True,
        output_field_list=[
            ResponseField(title="test1", type=int, required=True),
            ResponseField(title="test2", type=str, required=True),
        ],
        context=[sub_task]
    )
    res = main_task.execute_sync(agent=agent, context="plan a Black Friday campaign.")

    assert isinstance(res, TaskOutput)
    assert res.task_id is main_task.id
    assert res.raw is not None
    assert isinstance(res.raw, str)
    assert res.json_dict is not None
    assert isinstance(res.json_dict, dict)
    assert res.pydantic is not None

    if hasattr(res.pydantic, "output"):
        assert res.pydantic.output is not None
    else:
        assert hasattr(res.pydantic, "test1")
        if res.pydantic.test1:
            assert type(res.pydantic.test1) == int | str

        assert hasattr(res.pydantic, "test2")
        if res.pydantic.test2:
            assert type(res.pydantic.test2) == list | str

    assert sub_task.output is not None
    assert sub_task.output.json_dict is not None
    assert sub_task.output.pydantic is None

    assert "result" in main_task.prompt()
    assert main_task.prompt_context == "plan a Black Friday campaign."
    assert "plan a Black Friday campaign." in main_task.prompt()


def test_callback():
    """
    See if the callback function is executed well with kwargs.
    """

    def callback_func(kwargs: Dict[str, Any]):
        task_id = kwargs.get("task_id", None)
        added_condition = kwargs.get("added_condition", None)
        assert task_id is not None
        assert added_condition is not None
        return f"Result: {task_id}, condition added: {added_condition}"

    agent = Agent(role="demo agent 5", goal="My amazing goals")
    task = Task(
        description="Analyze the client's business model and define the optimal cohort timeframe.",
        expected_output_json=True,
        expected_output_pydantic=False,
        output_field_list=[
            ResponseField(title="test1", type=str, required=True),
        ],
        callback=callback_func,
        callback_kwargs=dict(added_condition="demo for pytest")
    )

    res = task.execute_sync(agent=agent)

    assert res is not None
    assert isinstance(res, TaskOutput)
    assert res.task_id is task.id
    assert res.raw is not None


def test_delegate():
    agent = Agent(role="demo agent 6", goal="My amazing goals")
    task = Task(
        description="Analyze the client's business model and define the optimal cohort timeframe.",
        expected_output_json=True,
        expected_output_pydantic=False,
        output_field_list=[
            ResponseField(title="test1", type=str, required=True),
        ],
        allow_delegation=True
    )

    task.execute_sync(agent=agent)

    assert task.output is not None
    assert "delegated_agent" in task.processed_by_agents
    assert task.delegations != 0


def test_conditional_task():
    agent = Agent(role="demo agent 6", goal="My amazing goals")
    task = Task(
        description="Analyze the client's business model and define the optimal cohort timeframe.",
        output_field_list=[ResponseField(title="test1", type=str, required=True),],
    )
    res = task.execute_sync(agent=agent)

    conditional_task = ConditionalTask(
        description="Analyze the client's business model.",
        output_field_list=[ResponseField(title="test1", type=str, required=True),],
        condition=lambda x: bool("zzz" in task.output.raw)
    )
    should_execute = conditional_task.should_execute(context=res)

    assert res.raw is not None
    assert should_execute is False

    conditional_res = conditional_task._handle_conditional_task(task_outputs=[res,], task_index=1, was_replayed=False)
    if not should_execute:
        assert conditional_res is None
    else:
        assert conditional_res.task_id is conditional_task.id


def test_store_task_log():
    task = Task(
        description="Analyze the client's business model.",
        output_field_list=[ResponseField(title="task_1", type=str, required=True),],
    )
    assert task._task_output_handler.load() is not None


def test_task_with_agent_tools():
    def empty_func():
        return "empty function"

    class CustomTool(Tool):
        name: str = "custom tool"

        def _run(self) -> str:
            return "empty function"

    custom_tool = CustomTool()

    agent_str_tool = Agent(role="demo 1", goal="test a tool", tools=["random tool 1", "random tool 2",])
    agent_dict_tool = Agent(role="demo 2", goal="test a tool", tools=[dict(name="tool 1", function=empty_func)])
    agent_custom_tool = Agent(role="demo 3", goal="test a tool", tools=[custom_tool, custom_tool])
    agents = [agent_str_tool, agent_dict_tool, agent_custom_tool]

    task = Task(description="execute the function", can_use_agent_tools=True, take_tool_res_as_final=True)

    for agent in agents:
        res = task.execute_sync(agent=agent)
        assert "empty function" in res.tool_output
        assert len(res.tool_output) == len(agent.tools)


def test_task_with_tools():

    def random_func(str: str = None) -> str:
        return str

    class CustomTool(Tool):
        name: str = "custom tool"

        def _run(self) -> str:
            return "custom function"

    tool = Tool(name="tool", function=random_func)
    custom_tool = CustomTool()
    tool_set = ToolSet(tool=tool, kwargs=dict(str="empty function"))
    task = Task(description="execute the function", tools=[custom_tool, tool, tool_set], take_tool_res_as_final=True)
    agent = Agent(role="demo", goal="demo")
    res = task.execute_sync(agent=agent)

    assert res.tool_output is not None
    assert isinstance(res.tool_output, list)
    assert len(res.tool_output) == len(task.tools)
    assert res.tool_output[0] == "custom function"
    assert res.tool_output[1] is None
    assert res.tool_output[2] == "empty function"


def test_create_json_dict_output():
    agent = Agent(role="demo agent 6", goal="My amazing goals")
    task = Task(
        description="Analyze the client's business model.",
        output_field_list=[
            ResponseField(title="str", type=str, required=True),
            ResponseField(title="list", type=list),
            ResponseField(title="dict", type=dict)
        ],
    )

    res = task.execute_sync(agent=agent)

    assert isinstance(res, TaskOutput)
    assert isinstance(res.json_dict, dict)
    assert [k == "str" for k, v in res.json_dict.items()]
    assert [k == "list" for k, v in res.json_dict.items()]
    assert [k == "dict" for k, v in res.json_dict.items()]
    assert isinstance(res.json_dict["str"], str)
    assert isinstance(res.json_dict["list"], list)
    assert isinstance(res.json_dict["dict"], dict)


# token usage logic

if __name__ == "__main__":
    test_create_json_dict_output()



import appdirs
import os
import json
import sqlite3
import datetime
from typing import Any, Dict, List, Optional
from dotenv import load_dotenv
from pathlib import Path

from versionhq._utils.logger import Logger

load_dotenv(override=True)

def fetch_db_storage_path():
    directory_name = Path.cwd().name
    data_dir = Path(appdirs.user_data_dir(appname=directory_name, appauthor="Version IO Sdn Bhd.", version=None, roaming=False))
    data_dir.mkdir(parents=True, exist_ok=True)
    return data_dir

storage_path = fetch_db_storage_path()
default_db_name = "task_outputs"


class TaskOutputSQLiteStorage:
    """
    An SQLite storage class to handle storing task outputs.
    """

    def __init__(self, db_path: str = f"{storage_path}/{default_db_name}.db") -> None:
        self.db_path = db_path
        self._logger = Logger(verbose=True)
        self._initialize_db()


    def _initialize_db(self):
        """
        Initializes the SQLite database and creates LTM table.
        """

        # try:
        with sqlite3.connect(self.db_path) as conn:
            cursor = conn.cursor()
            cursor.execute(
                """
                CREATE TABLE IF NOT EXISTS task_outputs (
                    task_id TEXT PRIMARY KEY,
                    output JSON,
                    task_index INTEGER,
                    inputs JSON,
                    was_replayed BOOLEAN,
                    timestamp DATETIME DEFAULT CURRENT_TIMESTAMP
                )
            """
            )
            conn.commit()

        # except sqlite3.Error as e:
        #     self._logger.log(level="error", message=f"DATABASE INITIALIZATION ERROR: {e}", color="red")


    def add(self, task, output: Dict[str, Any], task_index: int, was_replayed: bool = False, inputs: Dict[str, Any] = {}):
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute(
                        """INSERT OR REPLACE INTO task_outputs
                            (task_id, output, task_index, inputs, was_replayed, timestamp)
                            VALUES (?, ?, ?, ?, ?, ?)
                        """,
                     (str(task.id), json.dumps(output), task_index, json.dumps(inputs), was_replayed, datetime.datetime.now())
                )
                conn.commit()

        except sqlite3.Error as e:
            self._logger.log(level="error", message=f"SAVING TASK OUTPUTS ERROR: {e}", color="red")


    def update(self, task_index: int, **kwargs):
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                fields, values = [], []
                for k, v in kwargs.items():
                    fields.append(f"{k} = ?")
                    values.append(json.dumps(v) if isinstance(v, dict) else v)

                query = f"UPDATE latest_kickoff_task_outputs SET {', '.join(fields)} WHERE task_index = ?"
                values.append(task_index)
                cursor.execute(query, tuple(values))
                conn.commit()

                if cursor.rowcount == 0:
                    self._logger.log(
                        level="info", message=f"No row found with task_index {task_index}. No update performed.", color="yellow",
                    )

        except sqlite3.Error as e:
            self._logger.log(level="error", message=f"UPDATE TASK OUTPUTS ERROR: {e}", color="red")


    def load(self) -> Optional[List[Dict[str, Any]]]:
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute("""
                SELECT *
                FROM task_outputs
                ORDER BY task_index
                """)

                rows = cursor.fetchall()
                results = []
                for row in rows:
                    result = {
                        "task_id": row[0],
                        "output": json.loads(row[1]),
                        "task_index": row[2],
                        "inputs": json.loads(row[3]),
                        "was_replayed": row[4],
                        "timestamp": row[5],
                    }
                    results.append(result)
                return results

        except sqlite3.Error as e:
            self._logger.log(level="error", message=f"LOADING TASK OUTPUTS ERROR: {e}", color="red")
            return None


    def delete_all(self):
        try:
            with sqlite3.connect(self.db_path) as conn:
                cursor = conn.cursor()
                cursor.execute("DELETE FROM task_outputs")
                conn.commit()

        except sqlite3.Error as e:
            self._logger.log(level="error", message=f"ERROR: Failed to delete all: {e}", color="red")



import os
import pytest

from versionhq.agent.model import Agent
from versionhq.llm.model import LLM
from versionhq.tool.model import Tool

MODEL_NAME = os.environ.get("LITELLM_MODEL_NAME", "gpt-3.5-turbo")
LITELLM_API_KEY = os.environ.get("LITELLM_API_KEY")


def test_build_agent_with_minimal_input():
    agent = Agent(
        role="analyst",
        goal="analyze the company's website and retrieve the product overview",
    )

    assert agent.role == "analyst"
    assert agent.role in agent.backstory
    assert agent.goal in agent.backstory
    assert isinstance(agent.llm, LLM)
    assert agent.llm.model == "gpt-3.5-turbo"
    assert agent.llm.api_key == LITELLM_API_KEY
    assert agent.tools == []


def test_build_agent_with_backstory():
    agent = Agent(
        role="analyst",
        goal="analyze the company's website and retrieve the product overview",
        backstory="You are competitive analysts who have abundand knowledge in marketing, product management."
    )

    assert agent.role == "analyst"
    assert agent.backstory == "You are competitive analysts who have abundand knowledge in marketing, product management."
    assert isinstance(agent.llm, LLM)
    assert agent.llm.model == "gpt-3.5-turbo"
    assert agent.llm.api_key == LITELLM_API_KEY
    assert agent.tools == []


def test_build_agent():
    agent = Agent(
        role="analyst",
        goal="analyze the company's website and retrieve the product overview",
        knowledge="competitor products",
        skillsets=["financial analysis", "product management", ]
    )

    assert agent.role == "analyst"
    assert agent.role in agent.backstory
    assert agent.goal in agent.backstory
    assert agent.knowledge in agent.backstory
    for item in agent.skillsets:
        assert item in agent.backstory
    assert isinstance(agent.llm, LLM)
    assert agent.llm.model == "gpt-3.5-turbo"
    assert agent.llm.api_key == LITELLM_API_KEY
    assert agent.tools == []


def test_build_agent_with_llm():
    agent = Agent(
        role="analyst",
        goal="analyze the company's website and retrieve the product overview",
        knowledge="competitor products",
        skillsets=["financial analysis", "product management", ],
        llm="gpt-4o"
    )
    assert agent.role == "analyst"
    assert agent.role in agent.backstory
    assert agent.goal in agent.backstory
    assert agent.knowledge in agent.backstory
    for item in agent.skillsets:
        assert item in agent.backstory
    assert isinstance(agent.llm, LLM)
    assert agent.llm.model == "gpt-4o"
    assert agent.llm.api_key == LITELLM_API_KEY
    assert agent.tools == []


def test_agent_with_random_str_tools():
    agent = Agent(role="demo", goal="test a tool", tools=["random tool 1", "random tool 2",])

    assert [isinstance(tool, Tool) for tool in agent.tools]
    assert [tool.run() is not None for tool in agent.tools]
    assert [tool.run() == "empty function" for tool in agent.tools]
    assert agent.tools[0].name == "random tool 1"
    assert agent.tools[1].name == "random tool 2"


def test_agent_with_random_dict_tools():
    def empty_func():
        return "empty function"
    agent = Agent(role="demo", goal="test a tool", tools=[dict(name="tool 1", function=empty_func), ])

    assert [tool.run() == "empty function" for tool in agent.tools]
    assert agent.tools[0].name == "tool 1"


def test_agent_with_custom_tools():
    class CustomTool(Tool):
        name: str = "custom tool"

        def _run(self, agent_role: str) -> str:
            return agent_role

    tool = CustomTool()
    agent = Agent(role="demo", goal="test a tool", tools=[tool])

    assert agent.tools[0] is tool
    assert agent.tools[0].run(agent_role=agent.role) == "demo"
    assert agent.tools[0].name == "custom tool"



import os
import pytest
from unittest.mock import patch
from typing import Dict, Any, Callable, List

from versionhq.tool.model import Tool, ToolSet
from versionhq.tool.tool_handler import ToolHandler
from versionhq.tool.decorator import tool


def test_create_custom_tool():
    class CustomTool(Tool):
        name: str = "custom tool"
        goal: str = "test a custom tool class"

        def _run(self, sentence: str) -> str:
            return sentence

    tool = CustomTool()

    assert tool.name == "custom tool"
    assert tool.description == "Tool Name: custom tool\nTool Arguments: {'sentence': {'description': None, 'type': 'str'}}\nGoal: test a custom tool class"
    assert tool.args_schema.model_json_schema()["properties"] == {"sentence": { "title": "Sentence", "type": "string" }}
    assert tool.run(sentence="I am a sentence.") == "I am a sentence."


def test_run_tool_with_abstract_function():
    def function_to_test(test_list: List) -> int:
        return len(test_list)

    class CustomTool(Tool):
        name: str = "custom tool"
        goal: str = "test a custom tool class"

        def _run(self, test_list: List) -> int:
            return function_to_test(test_list)

    tool = CustomTool()
    result = tool.run(test_list=["demo 0", "demo 1", "demo 2"])

    assert tool.name is "custom tool"
    assert result == 3


def test_run_tool_with_function():
    def function_to_test(test_list: List) -> int:
        return len(test_list)

    tool = Tool(name="tool", goal="run a function", function=function_to_test)
    result = tool.run(test_list=["demo 0", "demo 1", "demo 2"])

    assert tool.name is "tool"
    assert result == 3


def test_setup_cache_function():
    class CustomTool(Tool):
        name: str = "custom tool"
        goal: str = "test a custom tool"
        cache_function: Callable = lambda: False

        def _run(self, sentence: str) -> str:
            return sentence

    tool = CustomTool()
    assert not tool.cache_function()


def test_default_cache_function():
    class CustomTool(Tool):
        name: str = "custom tool"
        goal: str = "test a custom tool"

        def _run(self, sentence: str) -> str:
            return sentence

    tool = CustomTool()
    assert tool.cache_function()


def test_tool_annotation():
    @tool("demo")
    def my_tool(test_words: str) -> str:
        """Test a tool decorator."""
        return test_words

    assert my_tool.name == "demo"
    assert my_tool.description == "Tool Name: demo\nTool Arguments: {'test_words': {'description': None, 'type': 'str'}}\nGoal: None"
    assert my_tool.args_schema.model_json_schema()["properties"] == { "test_words": {"title": "Test Words", "type": "string"}}
    assert my_tool.function("testing") == "testing"


def test_tool_handler_turning_off_cache():
    def empty_func():
        return "empty function"

    tool = Tool(name="demo", function=empty_func, should_cache=False)
    tool.run()

    assert isinstance(tool.tool_handler, ToolHandler)
    assert tool.tool_handler.should_cache == False
    assert tool.tool_handler.last_used_tool is None


def test_tool_handler_with_cache():
    def empty_func():
        return "empty function"

    tool = Tool(name="demo", function=empty_func, should_cache=True)
    tool.run()

    assert isinstance(tool.tool_handler, ToolHandler)
    assert tool.tool_handler.should_cache == True
    assert tool.tool_handler.last_used_tool == ToolSet(tool=tool, kwargs={})


if __name__ == "__main__":
    test_tool_handler_with_cache()



from abc import ABC, abstractmethod
from inspect import signature
from typing import Any, Dict, Callable, Type, Optional, get_args, get_origin
from typing_extensions import Self
from pydantic import InstanceOf, BaseModel, ConfigDict, Field, field_validator, model_validator

from versionhq._utils.cache_handler import CacheHandler


class BaseTool(ABC, BaseModel):
    """
    Abstract class for Tool class.
    """

    class _ArgsSchemaPlaceholder(BaseModel):
        pass

    args_schema: Type[BaseModel] = Field(default_factory=_ArgsSchemaPlaceholder)


    @field_validator("args_schema", mode="before")
    @classmethod
    def _default_args_schema(cls, v: Type[BaseModel]) -> Type[BaseModel]:
        if not isinstance(v, cls._ArgsSchemaPlaceholder):
            return v

        return type(
            f"{cls.__name__}Schema",
            (BaseModel,),
            { "__annotations__": { k: v for k, v in cls._run.__annotations__.items() if k != "return" }},
        )


    @abstractmethod
    def _run(self, *args: Any, **kwargs: Any,) -> Any:
        """any handling"""



class Tool(BaseTool):
    name: str = Field(default=None)
    goal: str = Field(default=None)
    function: Callable = Field(default=None)
    tool_handler: Optional[Dict[str, Any] | Any] = Field(default=None, description="store tool_handler to record the usage of this tool")
    should_cache: bool = Field(default=True, description="whether the tool usage should be cached")
    cache_function: Callable = lambda _args=None, _result=None: True
    cache_handler: Optional[InstanceOf[CacheHandler]] = Field(default=None)


    @model_validator(mode="after")
    def set_up_tool_handler(self) -> Self:
        from versionhq.tool.tool_handler import ToolHandler

        if self.tool_handler and not isinstance(self.tool_handler, ToolHandler):
            ToolHandler(**self.tool_handler)

        else:
            self.tool_handler = ToolHandler(cache_handler=self.cache_handler, should_cache=self.should_cache)

        return self


    @model_validator(mode="after")
    def set_up_function(self) -> Self:
        if self.function is None:
            self.function = self._run
            self._set_args_schema_from_func()
        return self


    @staticmethod
    def _get_arg_annotations(annotation: type[Any] | None) -> str:
        if annotation is None:
            return "None"

        origin = get_origin(annotation)
        args = get_args(annotation)

        if origin is None:
            return annotation.__name__ if hasattr(annotation, "__name__") else str(annotation)

        if args:
            args_str = ", ".join(Tool._get_arg_annotations(arg) for arg in args)
            return f"{origin.__name__}[{args_str}]"

        return origin.__name__


    def _parse_args(self, raw_args: str | dict) -> dict:
        """
        Parse and validate the input arguments against the schema
        """
        if isinstance(raw_args, str):
            try:
                import json
                raw_args = json.loads(raw_args)
            except json.JSONDecodeError as e:
                raise ValueError(f"Failed to parse arguments as JSON: {e}")

        try:
            validated_args = self.args_schema.model_validate(raw_args)
            return validated_args.model_dump()

        except Exception as e:
            raise ValueError(f"Arguments validation failed: {e}")


    def _set_args_schema_from_func(self):
        class_name = f"{self.__class__.__name__}Schema"
        self.args_schema = type(
            class_name,
            (BaseModel,),
            { "__annotations__": {
                k: v for k, v in self._run.__annotations__.items() if k != "return"
            } },
        )
        return self


    def _run(self, *args: Any, **kwargs: Any) -> Any:
        return self.run(*args, **kwargs)


    def run(self, *args, **kwargs) -> Any:
        """
        Use tool and record its usage if should_cache is True.
        """
        from versionhq.tool.tool_handler import ToolHandler

        result = None
        tool_set = ToolSet(tool=self, kwargs={})

        if self.function:
            result = self.function(*args, **kwargs)

        else:
            acceptable_args = self.args_schema.model_json_schema()["properties"].keys()
            acceptable_kwargs = { k: v for k, v in kwargs.items() if k in acceptable_args }
            tool_set = ToolSet(tool=self, kwargs=acceptable_kwargs)

            if self.tool_handler:
                if self.tool_handler.has_called_before(tool_set):
                    self.tool_handler.error = "Agent execution error"

                elif self.tool_handler.cache:
                    result = self.tools_handler.cache.read(tool=tool_set.tool.name, input=tool_set.kwargs)
                    if result is None:
                        parsed_kwargs = self._parse_args(raw_args=acceptable_kwargs)
                        result = self.function(**parsed_kwargs) if self.function else None

            else:
                tool_handler = ToolHandler(last_used_tool=tool_set, cache_handler=self.cache_handler, should_cache=self.should_cache)
                self.tool_handler = tool_handler
                parsed_kwargs = self._parse_args(raw_args=acceptable_kwargs)
                result = self.function(**parsed_kwargs) if self.function else None


        if self.should_cache is True:
            self.tool_handler.record_last_tool_used(tool_set, result, self.should_cache)

        return result


    @property
    def description(self) -> str:
        args_schema = {
            name: {
                "description": field.description,
                "type": self._get_arg_annotations(field.annotation),
            }
            for name, field in self.args_schema.model_fields.items()
        }

        return f"Tool Name: {self.name}\nTool Arguments: {args_schema}\nGoal: {self.goal}"


class ToolSet(BaseModel):
    """
    Store the tool called and any kwargs used.
    """
    tool: InstanceOf[Tool] | Any = Field(..., description="store the tool instance to be called.")
    kwargs: Optional[Dict[str, Any]] = Field(..., description="kwargs passed to the tool")


class InstructorToolSet(BaseModel):
    tool: InstanceOf[Tool] | Any = Field(..., description="store the tool instance to be called.")
    kwargs: Optional[Dict[str, Any]] = Field(..., description="kwargs passed to the tool")


class CacheTool(BaseModel):
    """
    Default tools to hit the cache.
    """

    name: str = "Hit Cache"
    cache_handler: CacheHandler = Field(default_factory=CacheHandler)

    def hit_cache(self, key):
        split = key.split("tool:")
        tool = split[1].split("|input:")[0].strip()
        tool_input = split[1].split("|input:")[1].strip()
        return self.cache_handler.read(tool, tool_input)



from typing import Optional, Any
from pydantic import InstanceOf

from versionhq.tool.model import ToolSet, InstructorToolSet
from versionhq._utils.cache_handler import CacheHandler


class ToolHandler:
    """
    Record the tool usage by ToolSet instance with cache and error recording.
    """

    last_used_tool: InstanceOf[ToolSet] | InstanceOf[InstructorToolSet]
    cache: Optional[CacheHandler]
    error: Optional[str]
    should_cache: bool

    def __init__(
        self,
        last_used_tool: InstanceOf[ToolSet] | InstanceOf[InstructorToolSet] = None,
        cache_handler: Optional[CacheHandler] = None,
        should_cache: bool = True
    ):
        self.cache = cache_handler
        self.last_used_tool = last_used_tool
        self.should_cache = should_cache


    def record_last_tool_used(
        self, last_used_tool: InstanceOf[ToolSet] | InstanceOf[InstructorToolSet], output: str, should_cache: bool = True
    ) -> Any:
        from versionhq.tool.model import CacheTool

        self.last_used_tool = last_used_tool

        if self.cache and should_cache and last_used_tool.tool.name != CacheTool().name:
            self.cache.add(tool=last_used_tool.tool.name, input=last_used_tool.kwargs, output=output)


    def has_called_before(self, tool_set: ToolSet = None) -> bool:
        if tool_set is None or not self.last_used_tool:
            return False

        if tool_set := self.last_used_tool:
            return bool((tool_set.tool.name == self.last_used_tool.tool.name) and (tool_set.kwargs == self.last_used_tool.kwargs))
